/**
 * Options for recording. All optional.
 */
export type RecorderOptions = {
    /**
     * A name for the recorder, used as prefix for the default file name.
     */
    name?: string;
    /**
     * The recording duration in seconds. If set to Infinity, `await canvasRecorder.stop()` needs to be called manually.
     */
    duration?: number;
    /**
     * The frame rate in frame per seconds. Use `await canvasRecorder.step();` to go to the next frame.
     */
    frameRate?: number;
    /**
     * Automatically download the recording when duration is reached or when `await canvasRecorder.stop()` is manually called.
     */
    download?: boolean;
    /**
     * Default file extension: infers which Encoder is selected.
     */
    extension?: boolean;
    /**
     * Default writing target: in-browser or file-system when available.
     */
    target?: string;
    /**
     * A specific encoder. Default encoder based on options.extension: GIF > WebCodecs > H264MP4.
     */
    encoder?: any;
    /**
     * See `src/encoders` or individual packages for a list of options.
     */
    encoderOptions?: any;
    onStatusChange?: onStatusChangeCb;
};
/**
 * Options for recording. All optional.
 */
export type RecorderStartOptions = {
    /**
     * Overwrite the file name completely.
     */
    filename?: string;
};
/**
 * A callback to notify on the status change. To compare with RecorderStatus enum values.
 * @name onStatusChangeCb
 * @function
 * @param {number} RecorderStatus the status
 */
/**
 * @typedef {Object} RecorderOptions Options for recording. All optional.
 * @property {string} [name=""] A name for the recorder, used as prefix for the default file name.
 * @property {number} [duration=10] The recording duration in seconds. If set to Infinity, `await canvasRecorder.stop()` needs to be called manually.
 * @property {number} [frameRate=30] The frame rate in frame per seconds. Use `await canvasRecorder.step();` to go to the next frame.
 * @property {boolean} [download=true] Automatically download the recording when duration is reached or when `await canvasRecorder.stop()` is manually called.
 * @property {boolean} [extension="mp4"] Default file extension: infers which Encoder is selected.
 * @property {string} [target="in-browser"] Default writing target: in-browser or file-system when available.
 * @property {Object} [encoder] A specific encoder. Default encoder based on options.extension: GIF > WebCodecs > H264MP4.
 * @property {Object} [encoderOptions={}] See `src/encoders` or individual packages for a list of options.
 * @property {onStatusChangeCb} [onStatusChange]
 */
/**
 * @typedef {Object} RecorderStartOptions Options for recording. All optional.
 * @property {string} [filename] Overwrite the file name completely.
 */
/**
 * Base Recorder class.
 * @property {boolean} [enabled=true] Enable/disable pointer interaction and drawing.
 */
export class Recorder {
    static defaultOptions: {
        name: string;
        duration: number;
        frameRate: number;
        download: boolean;
        extension: string;
        onStatusChange: () => void;
    };
    static mimeTypes: {
        mkv: string;
        webm: string;
        mp4: string;
        gif: string;
    };
    /**
     * @param {RenderingContext} context
     * @param {RecorderOptions} options
     */
    constructor(context: RenderingContext, options?: RecorderOptions);
    set width(arg: any);
    get width(): any;
    set height(arg: any);
    get height(): any;
    get stats(): {
        renderTime: number;
        secondsPerFrame: number;
        detail: string;
    };
    status: any;
    getParamString(): string;
    getDefaultFileName(extension: any): string;
    getSupportedExtension(): any;
    getSupportedTarget(): any;
    context: RenderingContext;
    encoder: WebCodecsEncoder | H264MP4Encoder | GIFEncoder | FrameEncoder;
    /**
     * Sets up the recorder internals and the encoder depending on supported features.
     * @private
     */
    private init;
    deltaTime: number;
    time: number;
    frame: number;
    frameTotal: number;
    /**
     * Start the recording by initializing and calling the initial step.
     * @param {RecorderStartOptions} startOptions
     */
    start({ filename }?: RecorderStartOptions): Promise<void>;
    startTime: Date;
    filename: string;
    /**
     * Convert the context into something encodable (bitmap, blob, buffer...)
     * @private
     */
    private getFrame;
    /**
     * Encode a frame and increment the time and the playhead.
     * Calls `await canvasRecorder.stop()` when duration is reached.
     */
    step(): Promise<void>;
    /**
     * Stop the recording and return the recorded buffer.
     * If options.download is set, automatically start downloading the resulting file.
     * Is called when duration is reached or manually.
     */
    stop(): Promise<any>;
    /**
     * Clean up
     */
    dispose(): Promise<void>;
    #private;
}
/**
 * *
 */
export type RecorderStatus = number;
/**
 * Enum for recorder status
 * @readonly
 * @enum {number}
 *
 * @example
 * ```js
 * // Check recorder status before continuing
 * if (canvasRecorder.status !== RecorderStatus.Stopped) {
 *   rAFId = requestAnimationFrame(() => tick());
 * }
 * ```
 */
export const RecorderStatus: Readonly<{
    Ready: 0;
    Initializing: 1;
    Recording: 2;
    Stopping: 3;
    Stopped: 4;
}>;
import WebCodecsEncoder from "./encoders/WebCodecsEncoder.js";
import H264MP4Encoder from "./encoders/H264MP4Encoder.js";
import GIFEncoder from "./encoders/GIFEncoder.js";
import FrameEncoder from "./encoders/FrameEncoder.js";
